from tensorflow.keras.callbacks import Callback
import matplotlib.pyplot as plt
from matplotlib import ticker
import pandas as pd
from pathlib import Path
import json
from datetime import datetime
import logging

class PlotCallback(Callback):

    status_file = "plots/plotstatus.tmp"

    def __init__(self, name, add_histograms=False, plot_live=False):
        self.name = name
        self.add_histograms = add_histograms
        self.plot_live = plot_live

        self.setup_plot()
        self.train_acc = []
        self.val_acc = []
        self.train_loss = []
        self.val_loss = []
        self.lr = []
        self.epochs = []
        self.highest_val_acc = 0
        self.highest_val_acc_epoch = 0
        self.plotpath = Path('plots') / f"{name}_plot.csv"
        self.resultspath = Path('logs') / f"{name}_results.json"
        self.attempt = 0
        self.parameters = {}
        self.results = {}

    def setup_plot(self):
        if not self.plot_live:
            return
        self.fig = plt.figure()
        self.fig.canvas.header_visible = False
        
        if self.add_histograms:
            self.ax_loss = self.fig.add_subplot(222)
            self.ax_acc = self.fig.add_subplot(221)
            self.ax_hist = self.fig.add_subplot(223)
            self.ax_feat = self.fig.add_subplot(224)
        else:
            self.ax_loss = self.fig.add_subplot(122)
            self.ax_acc = self.fig.add_subplot(121)

        self.ax_loss.legend()
        self.ax_acc.legend()
        plt.ion()
        self.fig.show()
        self.fig.canvas.draw()
        plt.tight_layout()

    def add_epochs(self, epochs):
        self.total_epochs = len(self.epochs) + epochs
        if not self.plot_live:
            return
        self.ax_loss.set_xlim([1,self.total_epochs])
        self.ax_loss.set_ylim([0,10])
        self.ax_acc.set_xlim([1,self.total_epochs])
        self.ax_acc.set_ylim([0,1])

    def on_epoch_end(self, epoch, logs=None):
        self.epochs.append(len(self.epochs)+1)
        self.train_acc.append(logs['accuracy'])
        self.train_loss.append(logs['loss'])
        self.val_acc.append(logs['val_accuracy'])
        self.val_loss.append(logs['val_loss'])
        self.lr.append(logs['lr_metric'])
        self.update_plot()
        self.save(logs)
        logging.info(f"Epoch #{epoch} val_acc = {logs['val_accuracy']:.4f} val_loss = {logs['val_loss']:.4f} lr = {logs['lr_metric']:.2E}")

    def save(self, logs):
        df = pd.DataFrame(list(zip(self.epochs, 
                self.train_acc, 
                self.train_loss, 
                self.val_acc, 
                self.val_loss,
                self.lr)), 
            columns = ["epoch", "accuracy", 'loss', 'val_accuracy', 'val_loss', 'lr_metric'])
        df.to_csv(self.plotpath, index=False)
        self.get_previous_attempts()
        result = {}
        result["name"] = self.parameters["name"]
        result["attempt"] = self.attempt
        result["datetime"] = datetime.now().strftime('%Y-%m-%d_%H:%M')
        result["last_epoch"] = len(self.epochs)
        result.update(self.parameters)
        result.update(logs)
        if self.highest_val_acc <= logs['val_accuracy']:
            self.highest_val_acc  = logs['val_accuracy']
            self.highest_val_acc_epoch = result['last_epoch']
        result['highest_val_acc'] = self.highest_val_acc
        result['highest_val_acc_epoch'] = self.highest_val_acc_epoch
        self.results[self.attempt] = result
        df: pd.DataFrame = pd.DataFrame.from_dict(self.results, orient='index')
        df.to_json(self.resultspath, orient='index')
        # fp = open(self.resultspath, 'w')
        # json.dump(self.results, fp)
        # fp.close()

    def get_previous_attempts(self):
        if self.resultspath.exists():
            df: pd.DataFrame = pd.read_json(self.resultspath, orient='index')
            self.results = df.to_dict(orient='index')
            # fp = open(self.resultspath, 'r')
            # self.results = json.load(fp)
            # fp.close()

    def update_plot(self):
        file = open(self.status_file, 'w')
        file.write(f"update,{str(self.plotpath)},{self.total_epochs}\n")
        file.close()
        if not self.plot_live:
            return
        plt.ion()
        self.ax_loss.clear()
        def plot(ax: plt.Axes, x, y, label):
            line, = ax.plot(x, self.smooth(y), label=label, lw=2)
            ax.plot(x, y, alpha=0.3, c=line.get_color())
        plot(self.ax_loss, self.epochs, self.train_loss, label='train_loss')
        plot(self.ax_loss, self.epochs, self.val_loss, label='val_loss')
        # self.ax_loss.plot(self.epochs, self.train_loss, label='train_loss')
        # self.ax_loss.plot(self.epochs, self.val_loss, label='val_loss')
        low = min(self.val_loss)
        high = max(max(self.train_loss), max(self.val_loss))
        ylim = min(high, low*20)
        low2 = min(low, min(self.train_loss))
        ylim_low = max(0, low2-2)
        self.ax_loss.set_xlim([1,self.total_epochs])
        self.ax_loss.set_ylim([ylim_low,ylim])
        self.ax_loss.grid(True, axis='y', which='major')
        self.ax_loss.grid(True, axis='y', which='minor', ls=":")
        self.ax_loss.legend()
        self.ax_acc.clear()
        plot(self.ax_acc, self.epochs, self.train_acc, 'train_acc')
        plot(self.ax_acc, self.epochs, self.val_acc, 'val_acc')
        # self.ax_acc.plot(self.epochs, self.train_acc, label='train_acc')
        # self.ax_acc.plot(self.epochs, self.val_acc, label='val_acc')
        self.ax_acc.set_xlim([1,self.total_epochs])
        self.ax_acc.set_ylim([0.4,1])
        self.ax_acc.get_yaxis().set_major_locator(ticker.MultipleLocator(0.1))
        self.ax_acc.get_yaxis().set_minor_locator(ticker.MultipleLocator(0.05))
        self.ax_acc.grid(True, axis='y', which='major')
        self.ax_acc.grid(True, axis='y', which='minor', ls=":")
        self.ax_acc.legend()
        if self.add_histograms:
            self.ax_feat.clear()
            self.ax_feat.hist(self.model.get_layer(name='final').get_weights()[0], bins=20)
        self.fig.show()
        self.fig.canvas.draw()

    def smooth(self, points, factor=0.8):
        smoothed_points = []
        for point in points:
            if smoothed_points:
                previous = smoothed_points[-1]
                smoothed_points.append(previous * factor + point * (1 - factor))
            else:
                smoothed_points.append(point)
        return smoothed_points

    def final_plot(self, y_pred):
        if self.add_histograms:
            self.ax_hist.hist(y_pred, bins=20)
