import numpy as np
from pathlib import Path
import nibabel as nib
import logging

class Pre_Data_Slices:

    def __init__(self, slice_shape, ids, labels):
        self.slice_shape = slice_shape
        self.ids = ids
        self.labels = labels
        self.make_slice_data_pre()
            
    def set_labels(self, i):
        self.y[i] = self.labels[i]
        self.y[i + self.total_ids] = self.labels[i]
        self.y[i + (self.total_ids * 2)] = self.labels[i]

    def store_index(self, i, pt_id):
        self.slice_indices[pt_id] = [i, i + self.total_ids, i + (self.total_ids * 2)]

    def collect_slices(self, i, pt_id):
        brain = list((Path("ADNI") / pt_id).rglob("*brain_reg2_norm.nii.gz"))[0]
        logging.info(f"Loading subject #{i}: {pt_id}")
        brain_data = nib.load(brain).get_fdata(dtype=np.float32)

        selected_slice_mid = brain_data.shape[2] // 2
        selected_slice_low = (brain_data.shape[2] // 2) - 15
        selected_slice_high = (brain_data.shape[2] // 2) + 15

        brain_axial_slice_mid = brain_data[:,:,selected_slice_mid]
        brain_axial_slice_low = brain_data[:,:,selected_slice_low]
        brain_axial_slice_high = brain_data[:,:,selected_slice_high]

        axial_shape_mid = brain_axial_slice_mid.shape
        axial_shape_low = brain_axial_slice_low.shape
        axial_shape_high = brain_axial_slice_high.shape

        brain_data_3d_mid = brain_axial_slice_mid.reshape(axial_shape_mid + (1,)) # Add third dimension (color)
        brain_data_3d_low = brain_axial_slice_low.reshape(axial_shape_low + (1,))
        brain_data_3d_high = brain_axial_slice_high.reshape(axial_shape_high + (1,))

        brain_data_colors_mid = np.tile(brain_data_3d_mid, (1,1,3)) # Repeat (tile) color dimension
        brain_data_colors_low = np.tile(brain_data_3d_low, (1,1,3))
        brain_data_colors_high = np.tile(brain_data_3d_high, (1,1,3))

        # mid slice
        sagg_start_mid = (self.slice_shape[0]-axial_shape_mid[0]) // 2
        cor_start_mid = (self.slice_shape[1]-axial_shape_mid[1]) // 2

        # low slice
        sagg_start_low = (self.slice_shape[0]-axial_shape_low[0]) // 2
        cor_start_low = (self.slice_shape[1]-axial_shape_low[1]) // 2

        # high slice
        sagg_start_high = (self.slice_shape[0]-axial_shape_high[0]) // 2
        cor_start_high = (self.slice_shape[1]-axial_shape_high[1]) // 2

        self.X[i, sagg_start_mid : sagg_start_mid + axial_shape_mid[0], cor_start_mid : cor_start_mid + axial_shape_mid[1], :] = brain_data_colors_mid
        logging.info("mid")
        self.X[i + self.total_ids, sagg_start_low : sagg_start_low + axial_shape_low[0], cor_start_low : cor_start_low + axial_shape_low[1], :] = brain_data_colors_low
        logging.info("low")
        self.X[i + (self.total_ids * 2), sagg_start_high : sagg_start_high + axial_shape_high[0], cor_start_high : cor_start_high + axial_shape_high[1], :] = brain_data_colors_high
        logging.info("high")

    def make_slice_data_pre(self):
        self.total_ids = len(self.ids)
        self.y = np.zeros(self.total_ids * 3)
        self.slice_indices = {}
        self.X = np.zeros((self.total_ids,) + (self.slice_shape,) + (3,), dtype=np.float32) # figure out if the rgb shit is neccesary
        for i, pt_id in enumerate(self.ids):
            self.set_labels(i)
            self.store_index(i, pt_id)
            self.collect_slices(i, pt_id)

class Pre_Data_Volume:

    def __init__(self, brain_shape, baseline_ids, longitudinal_ids):
        self.brain_shape = brain_shape
        self.baseline_ids = baseline_ids
        self.longitudinal_ids = longitudinal_ids
        if len(baseline_ids) > len(longitudinal_ids):
            logging.error("baseline ids list is longer than longitudinal ids list")
            raise Exception("all id list must be of same length")
        elif len(baseline_ids) < len(longitudinal_ids):
            logging.error("baseline ids list is shorter than longitudinal ids list")
            raise Exception("all id list must be of same length")
        self.make_volume_data_pre()

    def collect_brain(self, image_id):
        logging.info(f"Image ID = {image_id}")
        brain = list(Path("ADNI").rglob(f"*_I{image_id}.nii.gz"))[0].parent / "brain_reg2_norm.nii.gz"
        brain_data = nib.load(brain).get_fdata(dtype=np.float32)
        return brain_data / 1000.0

    def make_baseline_data(self): # must run first
        self.X_base = np.zeros((self.total_ids,) + self.brain_shape + (1,), dtype=np.float32)
        for i, image_id in enumerate(self.baseline_ids):
            logging.info(f"Loading subject #{i}: {image_id}")
            self.X_base[i, :, :, :, 0] = self.collect_brain(image_id)

    def make_longitudinal_data(self): # must run second
        self.X_long = np.zeros((self.total_ids,) + self.brain_shape + (1,), dtype=np.float32)
        for i, image_id in enumerate(self.longitudinal_ids):
            logging.info(f"Loading subject #{i + self.total_ids}: {image_id}")
            self.X_long[i, :, :, :, 0] = self.collect_brain(image_id)

    def make_volume_data_pre(self):
        self.total_ids = len(self.baseline_ids)
        self.make_baseline_data()
        self.make_longitudinal_data()

class LoadVolumes(Pre_Data_Volume):

    def __init__(self, brain_shape, baseline_ids):
        self.brain_shape = brain_shape
        self.baseline_ids = baseline_ids
        self.make_volume_data_pre()

    def make_volume_data_pre(self):
        self.total_ids = len(self.baseline_ids)
        self.make_baseline_data()

class End_Data_Slice:

    def __init__(self, y, indices):
        self.y = y
        self.indices = indices
        self.make_slice_data_end()
            
    def make_slice_data_end(self):
        self.y_end = []
        for pt_id in self.indices:
            overall = 0
            for i in self.indices[pt_id]:
                overall += self.y[i]
            if overall < 1.5:
                self.y_end.append(0)
            else:
                self.y_end.append(1)
        
class End_Data_Volume:

    def __init__(self, y):
        self.y = y
        self.make_volume_data_end()

    def make_volume_data_end(self):
        self.y_end = []
        for pred in self.y:
            if pred < 0.5:
                self.y_end.append(0)
            else:
                self.y_end.append(1)